// BlockFileThrasherDlg.h : header file
//

#if !defined(AFX_BLOCKFILETHRASHERDLG_H__75BC4A59_EC2B_11D6_8B22_00065B874763__INCLUDED_)
#define AFX_BLOCKFILETHRASHERDLG_H__75BC4A59_EC2B_11D6_8B22_00065B874763__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

/////////////////////////////////////////////////////////////////////////////
// CBlockFileThrasherDlg dialog

#include "Thread.h"
#include "BlockFile.h"

#define SENTINEL -1 //Largest possible Key field

class CCList {
private:
  struct CCListNode {
    CCListNode(DWORD key=SENTINEL) {Next=this; Key=key;}
    CCListNode* Next;
    DWORD        Key;
  };
  void Insert(DWORD Key) {Insert(new CCListNode(Key));}
  void Insert(CCListNode* NodeToInsert) {
    DWORD SearchValue=NodeToInsert->Key;
    for(it=First; it->Next->Key<SearchValue; it=it->Next);
    NodeToInsert->Next=it->Next;
    it->Next=NodeToInsert;
  }

protected:
  CCListNode* First;
  CCListNode* it; //Iterator Points at the Node PRIOR to the one to use.

public:
  CCList() {First=it=new CCListNode;}
  virtual ~CCList() {Empty(); delete First;}
  void Add   (DWORD Key) {Insert(Key);}
  void Delete(DWORD Key) {if(Find(Key)) Delete();}
  void Delete() {
    if(IsEmpty()) return;
    CCListNode* NodePtr=it->Next;
    it->Next=it->Next->Next;
    delete NodePtr;
  }
  bool IsEmpty() const {return First==First->Next;}
  void Empty() {
    while(!IsEmpty()) {
      it=First->Next;
      First->Next=First->Next->Next;
      delete it;
  } }
  void Home()   {it=First;}
  DWORD GetKey() {return it->Next->Key;}
  bool Next()   {it=it->Next; return it->Next!=First;}
  bool Find(DWORD SearchValue) {
    for(it=First; it->Next->Key<SearchValue; it=it->Next);
    return (it->Next->Key==SearchValue) && (it!=it->Next);
  }
};

// Since this extra routine scans the whole file twice, the Thrashing will get slower as the file gets bigger...
class CTestBlockFile : public CBlockFile {
public:
  bool GetStats(DWORD& UsedBlocks, DWORD& UsedBytes, DWORD& FreeBlocks, DWORD& FreeBytes) {
    DWORD SafePos=Pos;
    if(!IsOpen()) return true;
    UsedBlocks=UsedBytes=FreeBlocks=FreeBytes=0; //Start at the first block in the file
    CCList Used, Freed;
    for(DWORD NextPos=sizeof(First); NextPos<File.GetLength(); NextPos+=sizeof(Node)+Node.Size+sizeof(EndNode)) {
      if(!Goto(NextPos)) {
TRACE("Read Garbage!!! %x\r\n", NextPos);
        return false;
      }
      if(Node.Flags & FREE) {Freed.Add(NextPos); ++FreeBlocks; FreeBytes+=Node.Size;}
      else                  { Used.Add(NextPos); ++UsedBlocks; UsedBytes+=Node.Size;}
    }
//While we're here, check the lists:
    if(SafePos && !Used.Find(SafePos)) {
      if(Freed.Find(SafePos))
TRACE("Pos is pointing at a Free Block!!!\r\n");
      else
TRACE("Pos not pointing at a Used Block!!!\r\n");
      return false;
    }
//TRACE("\r\n-------\r\nUsed:\r\n");
    DWORD LastPos=0;
    for(Goto(First.Used); Pos; Next()) {
//TRACE("%x\r\n", Pos);
      if(Node.Prev!=LastPos) {
TRACE("Bad Prev: %x\r\n", Node.Prev);
        return false;
      }
      if(Used.Find(Pos)) Used.Delete(Pos);
      else {
TRACE("Used List Corrupt: %x\r\n", Node.Prev);
        return false;
      }
      LastPos=Pos;
    }
    if(!Used.IsEmpty()) {
      Used.Home();
      while(Used.Next()) {
TRACE("Used List Lost Block: %x\r\n", Used.GetKey());
        Used.Delete();
      }
      return false;
    }
//TRACE("-------\r\nFree:\r\n");
    LastPos=0;
    for(Goto(First.Free); Pos; Next()) {
//TRACE("%x\r\n", Pos);
      if(Node.Prev!=LastPos) {
TRACE("Bad Prev: %x\r\n", Node.Prev);
        return false;
      }
      if(Freed.Find(Pos)) Freed.Delete(Pos);
      else {
TRACE("Free List Corrupt: %x\r\n", Node.Prev);
        return false;
      }
      LastPos=Pos;
    }
    if(!Freed.IsEmpty()) {
      Freed.Home();
      while(Freed.Next()) {
TRACE("Free List Lost Block: %x\r\n", Freed.GetKey());
        Freed.Delete();
      }
      return false;
    }
    Goto(SafePos); //Don't care if this fails
    return NextPos==File.GetLength();
  }
};

class CBlockFileThrasherDlg : public CDialog, public CThread {
// Construction
public:
	CBlockFileThrasherDlg(CWnd* pParent=NULL);	// standard constructor
  CTestBlockFile BlockFile;
  void Main() {
    if(!BlockFile.Create("Test.bf")) {SetDlgItemText(IDC_Blocks, "Can't open BlockFile"); return;}
    BlockFile.Alloc(1);
    DWORD a=BlockFile.Alloc(32); //A few for intricate analysis ;-)
    DWORD b=BlockFile.Alloc(128);
    DWORD c=BlockFile.Alloc(32);
    BlockFile.Free(b);
    DWORD d=BlockFile.Alloc(32);
    BlockFile.Free(a);
    BlockFile.Free(c);
    BlockFile.Free(d);
//    BlockFile.Close();
    DWORD Blocks=1;
    DWORD Allocs=1;
    DWORD Frees=0;
    DWORD Size=1;
    DWORD UsedBlocks=0;
    DWORD UsedBytes=0;
    DWORD FreeBlocks=0;
    DWORD FreeBytes=0;
    DWORD Opened=0;
    DWORD Closed=0;
    DWORD Homed=0;
    DWORD Defragged=0;
    CString S;
    bool Changed=false;
    while(!Abort) {
      switch((rand()>>3)&7) {
        case 0: {
          if(rand()&0xFF) break;
          SetDlgItemText(IDC_Last, "Opened");
          if(!BlockFile.Open("Test.bf")) {
            S.Format("Opened %u times. Failed to open BlockFile!!!", ++Opened);
            SetDlgItemText(IDC_Opened, S);
            LPVOID lpMsgBuf; // Default language:
            FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM, NULL, GetLastError(), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR)&lpMsgBuf, 0, NULL);
            ::MessageBox(NULL, (LPTSTR)lpMsgBuf, "Error Information", MB_OK|MB_ICONINFORMATION);
            LocalFree(lpMsgBuf);
            return;
          }
          S.Format("Opened %u times.", ++Opened);
          SetDlgItemText(IDC_Opened, S);
          Changed=true;
          break;
        }
        case 1: {
          if(rand()&0xFF) break;
          SetDlgItemText(IDC_Last, "Home");
          BlockFile.Home();
          S.Format("Homed %u times.", ++Homed);
          SetDlgItemText(IDC_Homed, S);
          break;
        }
        case 2: {
          if(rand()&0x3FF) break;
          SetDlgItemText(IDC_Last, "Close");
          BlockFile.Close();
          S.Format("Closed %u times.", ++Closed);
          SetDlgItemText(IDC_Closed, S);
          break;
        }
        case 3: {
          SetDlgItemText(IDC_Last, "Next");
          BlockFile.Next();
          break;
        }
        case 4: {
          if(rand()&0x2) break;
          SetDlgItemText(IDC_Last, "Prev");
          BlockFile.Prev();
          break;
        }
        case 5: {
          if(FreeBlocks<20) break;
          SetDlgItemText(IDC_Last, "Defragment");
          Abort=!BlockFile.Defragment();
          S.Format("Defragged %u times.", ++Defragged);
          SetDlgItemText(IDC_Defragged, S);
          Changed=true;
          break;
        }
        case 6: {
          if(Size>0x1000000) break;
          SetDlgItemText(IDC_Last, "Alloc");
          DWORD dw=rand()&0xFF;
          if(BlockFile.Alloc(dw)) {
            dw=BlockFile.GetSize();
            S.Format("Blocks=%u.", ++Blocks);
            SetDlgItemText(IDC_Blocks, S);
            S.Format("Size=%u.", Size+=dw);
            SetDlgItemText(IDC_Size, S);
            S.Format("Allocs=%u. Last was %u Bytes.", ++Allocs, dw);
            Changed=true;
          }else S.Format("Allocs=%u. Failed trying for %u Bytes.", ++Allocs, dw);
          SetDlgItemText(IDC_Allocs, S);
          break;
        }
        case 7: {
          if((Size<0x800000) && (rand()&0x3)) break;
          SetDlgItemText(IDC_Last, "Free");
          DWORD dw=BlockFile.GetSize();
          if(BlockFile.Free()) {
            S.Format("Blocks=%u.", Blocks ? --Blocks : 0);
            SetDlgItemText(IDC_Blocks, S);
            S.Format("Size=%u.", Size-=dw);
            SetDlgItemText(IDC_Size, S);
            S.Format("Frees=%u.", ++Frees);
            SetDlgItemText(IDC_Frees, S);
            Changed=true;
          }
          break;
        }
      }
      if(Changed) {
        Changed=false;
        if(!BlockFile.GetStats(UsedBlocks, UsedBytes, FreeBlocks, FreeBytes)
          || (Blocks!=UsedBlocks)
          || (Size!=UsedBytes))
          Abort=true;
        S.Format("Used=%u Blocks taking up %u Bytes.", UsedBlocks, UsedBytes);
        SetDlgItemText(IDC_Used, S);
        S.Format("Free=%u Blocks taking up %u Bytes.", FreeBlocks, FreeBytes);
        SetDlgItemText(IDC_Free, S);
      }
    }
  }

// Dialog Data
	//{{AFX_DATA(CBlockFileThrasherDlg)
	enum { IDD = IDD_BLOCKFILETHRASHER_DIALOG };
		// NOTE: the ClassWizard will add data members here
	//}}AFX_DATA

	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CBlockFileThrasherDlg)
	//}}AFX_VIRTUAL

// Implementation
protected:
	HICON m_hIcon;

	// Generated message map functions
	//{{AFX_MSG(CBlockFileThrasherDlg)
	virtual BOOL OnInitDialog();
	afx_msg void OnDestroy();
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

//{{AFX_INSERT_LOCATION}}
// Microsoft Visual C++ will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_BLOCKFILETHRASHERDLG_H__75BC4A59_EC2B_11D6_8B22_00065B874763__INCLUDED_)
